# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
测试用例基类模块
'''
from __future__ import absolute_import

import os
import sys
import json
import re
import threading
import traceback
import collections
import types
import six

from testbase.assertion import AssertionRewriter
from testbase.util import Singleton, ThreadGroupLocal, ThreadGroupScope, smart_text, get_last_frame_stack, getmembers
from testbase.testresult import EnumLogLevel, TestResultCollection, TestResultType
from testbase.conf import settings
from testbase.retry import Retry


class BaseTestCaseException(Exception):
    pass


# 后续需专门花时间去除TestCaseStatus和TestCasePriority这两个类
class TestCaseStatus(object):
    '''测试用例状态

    :attention: 此类将会被移除，请使用TestCase.EnumStatus
    '''
    Design, Implement, Review, Ready, Suspend = ('Design', 'Implement', 'Review', 'Ready', 'Suspend')


class TestCasePriority(object):
    '''测试用例优先级

    :attention: 此类将会被移除，请使用TestCase.EnumPriority
    '''
    BVT, High, Normal, Low = ('BVT', 'High', 'Normal', 'Low')


class Environ(six.with_metaclass(Singleton, dict)):
    """测试环境类

用法说明：
它是一个继承了字典类型的单实例类。使用时，需先实例化该类，如：
from testbase.testcase import Environ
env = Environ()
env保存了用例运行时的一些测试环境变量。

测试环境变量分为3个部分：

1、在测试执行时，env则存储了由测试计划定义的用例环境变量，使用方法如下::

    from testbase.testcase import Environ
    env = Environ()
    print(env_
    # 输出
    env = {
        'ASSERTTEST':'True',  #注意：这里key是全字母大写
    }

2、测试用例基类testbase.testcase.TestCase的构造函数中也实例化了Environ类（变量名为environ)，
并且，保存了当时执行用例类的类名和类说明，使用方法如下::

    #todo: 执行用例中使用evniron，用于打印用例名称和用例说明
    from testbase.testcase import TestCase
    class YourTest(TestCase):
        def runTest(self):
            print(self.environ['TestName'])
            print(self.environ['TestDoc'])

3、Environ还可以用于设置自定义的环境变量，使用方法如下::

    from testbase.testcase import Environ
    env = Environ()
    env['YourEnvKey'] = "EnvValue"
    print(env['YourEnvKey'])

    """

    def __init__(self):
        """构造函数。判断系统环境变量中是否存在由测试计划中传入的环境变量，有则加载。

        """
        super(Environ, self).__init__()
        for key in os.environ.keys():
            if key.startswith("QTAF_") and key[5:]:
                self[key[5:]] = os.environ[key]


class TestCaseType(type):
    '''测试用例元类型
    '''
    forbid_overload_methods = ["__init__"]

    def __new__(cls, name, bases, attrs):
        super_new = super(TestCaseType, cls).__new__
        parents = [b for b in bases if isinstance(b, TestCaseType)]
        # ensure only apply for subclass of TestCase
        if not parents:
            return super_new(cls, name, bases, attrs)

        for it in cls.forbid_overload_methods:
            if it in attrs.keys():
                raise RuntimeError("不允许%s重载函数: %s" % (cls.__name__, it))

        base_tags_set = set()
        tags = attrs.pop("tags") if "tags" in attrs else set()
        if isinstance(tags, six.string_types):
            tags = [tags]
        tags_set = set(tags)
        for b in bases:
            if issubclass(b, TestCase) and hasattr(b, "tags"):
                base_tags_set |= b.tags
        if "__module__" in attrs:
            mod = sys.modules[attrs["__module__"]]
            if hasattr(mod, "__qtaf_tags__"):
                mod_tags = mod.__qtaf_tags__
                if isinstance(mod_tags, six.string_types):
                    mod_tags = [mod_tags]
                base_tags_set |= set(mod_tags)
        tags_set |= base_tags_set

        attrs["tags"] = tags_set
        attrs["base_tags"] = base_tags_set
        return super_new(cls, name, bases, attrs)


@six.add_metaclass(TestCaseType)
class TestCase(object):
    '''测试用例基类

            所有测试用例都最终从此基类继承。测试用例的测试脚本主要实现在"runTest()"中，

           而当用例需要初始化和清理测试环境时则分别重写"preTest()"和"postTest()"函数。
    '''
    test_extra_info_def = []  # 自定义字段

    class EnumStatus(object):
        '''测试用例状态枚举类

        :attention: 如果因为特殊原因需要暂时屏蔽某个用例的任务执行（比如有功能缺陷从而导致执行失败），
                                                     则可以先置为该字段为Suspend,等到可用的时候再将该字段置为Ready

        '''
        Design, Implement, Review, Ready, Suspend = (TestCaseStatus.Design,
                                                     TestCaseStatus.Implement,
                                                     TestCaseStatus.Review,
                                                     TestCaseStatus.Ready,
                                                     TestCaseStatus.Suspend)

    class EnumPriority(object):
        '''测试用例优先级枚举类
        '''
        BVT, High, Normal, Low = (TestCasePriority.BVT,
                                  TestCasePriority.High,
                                  TestCasePriority.Normal,
                                  TestCasePriority.Low)

    owner = None
    priority = None
    status = None
    timeout = None

    ATTRIB_OVERWRITE_WHITELIST = ["priority", "status", "owner", "timeout", "tags", "__doc__"]
    EXTRA_PROPERTY_BLACKLIST = ["start_time", "end_time", "begintime", "endtime", "duration"]

    def __init__(self, testdata=None, testdataname=None, attrs=None):
        '''构造函数

        :param testdata: 测试数据
        :type testdata: object
        :param testdataname: 测试数据标识
        :type testdataname: str
        :param attrs: 重载的测试用例属性
        :type attrs: dict
        '''
        self.__casedata = testdata
        self.__testdataname = testdataname
        self.__environ = Environ()
        self.__testresult = None
        self.__resmgr_session = None
        self.__resmgr = None
        self.__share_data_mgr = None
        self.__test_doc = None

        # 参数相关
        self.__params_definitions = {}  # add_params原始的用例参数数据
        self.__params = {}  # key，name的用例参数数据
        self.__dynamic_params = {}  # 外部传入的参数数据

        if attrs:
            for k, v in attrs.items():
                if k in self.ATTRIB_OVERWRITE_WHITELIST:
                    if k == "__doc__":
                        self.__test_doc = v
                    elif k == "tags":
                        if isinstance(v, six.string_types):
                            v = [v]
                        self.tags = self.base_tags | set(v)
                    else:
                        setattr(self, k, v)
                else:
                    self.__dynamic_params[k] = v

        self.handle_params()

    @property
    def casedata(self):
        '''测试数据

        :rtype: list
        '''
        return self.__casedata

    @property
    def casedataname(self):
        '''测试数据标识

        :rtype: str
        '''
        return self.__testdataname

    @property
    def environ(self):
        '''环境变量

        :rtype: Environ
        '''
        return self.__environ

    @property
    def test_result(self):
        '''对应的测试结果

        :rtype: TestResult
        '''
        return self.__testresult

    @property
    def test_dir(self):
        '''测试用例执行的临时目录

        :rtype: str
        '''
        return os.path.abspath(os.getcwd())

    @property
    def test_class_name(self):
        '''返回测试用例名字（不同测试用例的名字不同）

        :rtype: str
        '''
        cls = type(self)
        if cls.__module__ == '__main__':
            type_name = smart_text(cls.__name__)
        else:
            type_name = smart_text(cls.__module__ + '.' + cls.__name__)
        return type_name

    @property
    def test_name(self):
        '''返回测试用例实例的名字

        :rtype: str
        '''
        if self.casedataname is not None:
            casedataname = smart_text(self.casedataname)
            return '%s/%s' % (self.test_class_name, casedataname)
        else:
            return self.test_class_name

    @property
    def test_doc(self):
        '''测试用例说明

        :rtype: str
        '''
        if self.__test_doc:
            return self.__test_doc
        desc = self.__class__.__doc__
#        if desc:
#            desc = cgi.escape(desc)
        if isinstance(desc, six.text_type):
            desc = re.sub('^\s*', '', desc)
            desc = re.sub('\s*$', '', desc)
        return desc

    @property
    def test_extra_info(self):
        '''测试用例额外信息
        '''
        info = {}
        for name, _ in self.test_extra_info_def:
            info[name] = getattr(self, name, None)
        return info

    @property
    def test_resmgr(self):
        '''资源管理器
        '''
        return self.__resmgr

    @test_resmgr.setter
    def test_resmgr(self, resmgr):
        self.__resmgr = resmgr

    @property
    def test_resources(self):
        '''资源管理使用接口
        '''
        if not self.__resmgr_session:
            self.__resmgr_session = self.__resmgr.create_session(self)
        return self.__resmgr_session

    @property
    def share_data_mgr(self):
        '''共享数据管理器
        '''
        return self.__share_data_mgr

    @share_data_mgr.setter
    def share_data_mgr(self, share_data_mgr):
        self.__share_data_mgr = share_data_mgr

    @property
    def params_definitions(self):
        return self.__params_definitions

    @property
    def qtaf_params(self):
        return self.__params

    @property
    def dynamic_params(self):
        return self.__dynamic_params

    def get_test_extra_properties(self):
        var_dicts = {}
        for k, v in getmembers(self):
            if k in self.ATTRIB_OVERWRITE_WHITELIST:
                continue
            if k in self.EXTRA_PROPERTY_BLACKLIST:
                continue
            if not (isinstance(v, bool) or isinstance(v, int) or isinstance(v, float) or isinstance(v, (six.string_types, six.binary_type))):
                continue
            if k.islower() and not k.startswith('_') and not k.startswith('test_'):
                if isinstance(v, (six.string_types, six.binary_type)):
                    v = smart_text(v)
                var_dicts[k] = v
        return var_dicts

    def add_share_data(self, name, value, level=0):
        '''添加共享数据

        :type name: string
        :param name: 需要共享的数据名称
        :param value: 需要共享的数据内容
        '''
        if self.share_data_mgr:
            self.share_data_mgr.set(name, value, level)

    def get_share_data(self, name):
        '''从内存中获取存储的全局数据，给当前用例使用'''
        if self.share_data_mgr:
            return self.share_data_mgr.get(name)

    def remove_share_data(self, name):
        '''从内存中移除存储的共享数据'''
        if self.share_data_mgr:
            return self.share_data_mgr.remove(name)

    def add_parameters(self):
        # 用例重写或者不重写
        self.add_params()

    def add_parameter(self, name, type=str, optional=False, default=None, description='', validation=None,
                      rewrite=False, **kwargs):
        if type not in (bool, list, str, int, float, dict, "json"):
            raise BaseTestCaseException('Unsupported type of parameter.')

        if rewrite is False:  # 如果不能重写，需要判断是否存在
            assert name not in self.params_definitions, \
                'Parameters with the same name "{}" were found.'.format(name)

        if not optional and default is None:
            raise BaseTestCaseException("Parameter {} must define default value".format(name))

        self.__params_definitions[name] = {
            'name': name,
            'type': type,
            'optional': optional,
            'default': default,
            'description': description,
            'validation': validation,
            'kwargs': kwargs
        }

    def add_params(self):
        # 用例重写
        pass

    add_param = add_parameter

    def assert_types(self):
        '''参数校验并设置self.qtaf_params
        '''
        errors = []
        for d_name in self.params_definitions:
            default = self.params_definitions[d_name]["default"]
            define_type = self.params_definitions[d_name]['type']
            optional = self.params_definitions[d_name]['optional']
            for c_name in self.dynamic_params:
                c_name = smart_text(c_name)
                if c_name == d_name:
                    # 如果有外部传入，直接进行参数类型检查，检查正确设置参数否则记录异常
                    value = self.dynamic_params[c_name]
                    if optional and value is None and default is None:
                        self.__params[c_name] = value
                        break
                    if isinstance(value, six.text_type):
                        value = smart_text(value)
                    if self.params_definitions[d_name]['type'] == "json":
                        # json单独处理
                        try:
                            value = json.loads(self.dynamic_params[c_name])
                            self.__params[c_name] = value
                        except Exception:
                            errors.append("Parameter:{}, value:{}, is not json type.".format(c_name, value))
                    else:
                        if not isinstance(value, define_type):
                            # 如果期望类型是float，int型的数据, 转换为字符串，判断是否是数值型数据，如果是强制转换为数值型数据
                            ret = self.translate2type(define_type, value)
                            if ret is not None:
                                self.__params[c_name] = ret
                            else:
                                errors.append(
                                    "Parameter:{}, value:{}, type {} is not:{} type.".format(c_name,
                                                                                             value,
                                                                                             type(value),
                                                                                             str(define_type)))
                        else:
                            self.__params[c_name] = value
                    break
            else:
                # 如果外部未传入使用默认参数
                if default is not None:
                    if define_type == "json":
                        # json单独处理
                        try:
                            value = json.loads(default)
                            default = value
                        except Exception:
                            raise BaseTestCaseException("Parameter {}, value {}, is not json type.".format(d_name,
                                                                                                           default))
                    else:
                        if not isinstance(default, define_type):
                            raise BaseTestCaseException(
                                "Parameter {}, value {}, is not {} type.".format(d_name, default,
                                                                                 repr(define_type.__name__)))
                self.__params[d_name] = default

        # validation检查
        for d_name in self.params_definitions:
            validation = self.params_definitions[d_name]['validation']
            kwargs = self.params_definitions[d_name]['kwargs']
            default = self.params_definitions[d_name]['default']
            if kwargs is None:
                kwargs = {}
            if not isinstance(kwargs, dict):
                errors.append("Parameter:{}, param(kwargs) must be dict.".format(d_name))
            if d_name not in self.qtaf_params:
                value = default
                # 如果是default就不进行校验
            else:
                value = self.qtaf_params[d_name]

            if value == default:  # 如果值与default相等就不进行校验
                continue
            # 否则进行校验
            if validation is not None and callable(validation):
                try:
                    validation(value, self.qtaf_params, **kwargs)
                except Exception as exe:
                    errors.append(
                        "Parameter:{}, value:{}, validation error, detail: {}.".format(d_name, value, str(exe)))
        if errors:  # 如果参数校验失败抛出异常
            raise BaseTestCaseException("\n".join(errors))

    def handle_params(self):
        '''
        加载并设置参数,用例中可重写加载方法方式; 并显示的调用以适配不支持参数传递的方式
        正常模式：不设置参数
        设置数据驱动参数：只设置数据驱动参数
        设置全部参数：如果传入参数，以传入参数为准，否则设置数据驱动参数
        '''
        # 通过参数控制是否进行参数传递
        param_mode = settings.get("QTAF_PARAM_MODE", False)
        if not param_mode:
            param_mode = settings.get("QC_PARAM_MODE", False)

        if not param_mode:  # 正常模式
            return

        self.add_params()  # 添加参数定义
        self.assert_types()  # 参数校验

        temp_data = self.qtaf_params

        if param_mode == "DATADRIVE":
            if isinstance(self.casedata, dict):
                temp_data.update(self.casedata)
        elif param_mode == "PARAM" or param_mode is True:
            if isinstance(self.casedata, dict):
                temp_data.update(self.casedata)

        if not temp_data or not isinstance(temp_data, dict):
            return
        for key, value in temp_data.items():
            if not isinstance(key, six.string_types):
                continue
            setattr(self, str(key), value)

    def init_test(self, testresult):
        '''初始化测试用例。慎用此函数，尽量将初始化放到preTest里。

        :param testresult: 测试结果
        :type testresult: TestResult
        '''
        self.__testresult = testresult

    def pre_test(self):
        '''测试环境初始化
        '''
        pass

    def run_test(self):
        '''运行测试用例
        '''
        raise NotImplementedError("请在%s类中实现runTest方法" % type(self))

    def post_test(self):
        '''测试环境清理
        '''
        pass

    def clean_test(self):
        '''测试用例反初始化。慎用此函数，尽量将清理放到postTest里。
        '''
        if self.__resmgr_session:
            self.test_resources.destroy()
            self.__resmgr_session = None

    def start_step(self, stepinfo):
        '''开始执行一个测试步骤

        :param stepinfo: 步骤描述
        :type stepinfo: str
        '''
        if not isinstance(stepinfo, six.text_type):
            stepinfo = str(stepinfo)
        self.__testresult.begin_step(stepinfo)

    def log_info(self, info):
        '''Log一条信息

        :type info: string
        :param info: 要Log的信息
        '''
        if not isinstance(info, six.text_type):
            info = str(info)
        self.__testresult.info(info)

    def fail(self, message):
        '''测试用例失败

        :type message: string
        :param message: 要Log的信息
        '''
        if not isinstance(message, six.text_type):
            message = str(message)
        self.__testresult.error(message)

    def __record_assert_failed(self, message, actual, expect):
        '''记录Assert失败信息

        :param message: 提示信息
        :type message: string
        :param actual: 实际值
        :type actual: string
        :param expect: 期望值
        :type expect: string
        '''
        # 得到上一个函数调用帧所在的文件路径，行号，函数名
        stack = get_last_frame_stack(3)
        msg = "检查点不通过\n%s%s\n期望值：%s%s\n实际值：%s%s" % (smart_text(stack), smart_text(message),
                                                    expect.__class__, expect,
                                                    actual.__class__, actual)
        self.__testresult.log_record(EnumLogLevel.ASSERT, msg)
        if not settings.get("QTAF_ASSERT_CONTINUE", True):
            raise RuntimeError("testcase assert failed:%s" % message)

    def _log_assert_failed(self, message, back_count=2):
        """记录断言失败的信息
        """
        stack = get_last_frame_stack(back_count)
        msg = "检查点不通过\n%s%s\n" % (smart_text(stack), smart_text(message))
        self.__testresult.log_record(EnumLogLevel.ASSERT, msg)
        if not settings.get("QTAF_ASSERT_CONTINUE", True):
            raise RuntimeError("testcase assert failed:%s" % message)

    def assert_(self, message, value):
        """测试断言，如果value的值不为真，则用例失败，输出对应信息

        :param message:断言失败时的提示消息
        :type  message: str
        :param value:用于判断的值
        :type  value: bool或
        """
        if not value:
            self._log_assert_failed(message, 3)

    def assert_equal(self, message, actual, expect=True):
        '''检查实际值和期望值是否相等，不能则测试用例失败

       :param message: 检查信息
       :param actual: 实际值
       :param expect: 期望值(默认：True)
       :return: True or False
        '''
        if isinstance(actual, six.string_types):
            actual = smart_text(actual)
        if isinstance(expect, six.string_types):
            expect = smart_text(expect)
        if expect != actual:
            self.__record_assert_failed(message, actual, expect)
            return False
        else:
            return True

    def assert_match(self, message, actual, expect):
        '''检查actual和expect是否模式匹配，不匹配则记录一个检查失败

        :type message: string
        :param message: 失败时记录的消息
        :type actual: string
        :param actual: 需要匹配的字符串
        :type expect: string
        :param expect: 要匹配的正则表达式
        :return: 匹配成果
        '''
        if isinstance(actual, six.string_types):
            actual = smart_text(actual)
        if isinstance(expect, six.string_types):
            expect = smart_text(expect)
        if re.search(expect, actual):
            return True
        else:
            self.__record_assert_failed(message, actual, expect)
            return False

    def wait_for_equal(self, message, obj, prop_name, expected, timeout=10, interval=0.5):
        '''每隔interval检查obj.prop_name是否和expected相等，如果在timeout时间内都不相等，则测试用例失败

        :param message: 失败时的输出信息
        :param obj: 需要检查的对象
        :type prop_name: string
        :param prop_name: 需要检查的对象的属性名，支持多层属性
        :param expected: 期望的obj.prop_name值
        :param timeout: 超时秒数
        :param interval: 重试间隔秒数
        :return: True or False
        '''
        for _ in Retry(timeout=timeout, interval=interval, raise_error=False):
            actual = getattr(obj, prop_name)
            if actual == expected:
                return True
        else:
            self.__record_assert_failed(message, getattr(obj, prop_name), expected)
            return False

    def wait_for_match(self, message, obj, prop_name, expected, timeout=10, interval=0.5):
        '''每隔interval检查obj.prop_name是否和正则表达式expected是否匹配，如果在timeout时间内都不相等，则测试用例失败

        :param message: 失败时的输出信息
        :param obj: 需要检查的对象
        :type prop_name: string
        :param prop_name: 需要检查的对象的属性名, obj.prop_name返回字符串
        :param expected: 需要匹配的正则表达式
        :param timeout: 超时秒数
        :param interval: 重试间隔秒数
        :return: True or False
        '''
        for _ in Retry(timeout=timeout, interval=interval, raise_error=False):
            actual = getattr(obj, prop_name)
            if re.match(expected, actual, re.I):
                return True
        else:
            self.__record_assert_failed(message, getattr(obj, prop_name), expected)
            return False

    def get_extra_fail_record(self):
        '''当错误发生时，获取需要额外添加的日志记录和附件信息

        :rtype: dict,dict - 日志记录，附件信息
        '''
        return {}, {}

    def debug_run(self):
        '''本地调试测试用例
        '''
        from testbase import datadrive
        from testbase.runner import TestRunner
        from testbase.report import StreamTestReport, EmptyTestReport
        from testbase.testresult import StreamResult

        test_cls = type(self)
        if datadrive.is_datadrive(test_cls) or settings.DATA_DRIVE:  # datadrvie
            if self.casedataname is not None:
                tests = datadrive.load_datadrive_tests(test_cls, self.casedataname)
                report = EmptyTestReport(lambda tc: StreamResult())
            else:
                tests = datadrive.load_datadrive_tests(test_cls)
                report = StreamTestReport(output_testresult=True, output_summary=True)
        else:
            tests = [self]
            report = EmptyTestReport(lambda tc: StreamResult())
        runner = TestRunner(report)
        return runner.run(tests)

    def debug_run_one(self, name=None):
        '''本地调试测试用例，给数据驱动的用例使用，只执行一个用例

        :param name: 测试数据名称，如果不指定，执行第一个数据的用例
        '''
        from testbase import datadrive
        from testbase.runner import TestRunner
        from testbase.report import EmptyTestReport
        from testbase.testresult import StreamResult

        test_cls = type(self)
        if not datadrive.is_datadrive(test_cls) and not settings.DATA_DRIVE:  # non-datadrive
            raise RuntimeError("非数据驱动用例，请使用debug_run进行调试")
        if name:
            tests = datadrive.load_datadrive_tests(test_cls, name)
        else:
            tests = datadrive.load_datadrive_tests(test_cls)
            tests = tests[:1]
        runner = TestRunner(EmptyTestReport(lambda tc: StreamResult()))
        return runner.run(tests)

    #----------------------------------------------------
    #    以下为兼容老的代码风格的接口，新代码请勿再使用
    #----------------------------------------------------

    def run(self):
        '''本地调试测试用例
        '''
        self.debug_run()

        sys.stderr.write('============================================================\n')
        sys.stderr.write('注意：TestCase.run接口准备废弃，建议改用为.debug_run接口，例如：\n')
        sys.stderr.write('if __name__ == \'__main__\'\n')
        sys.stderr.write('    %s().debug_run()\n' % type(self).__name__)
        sys.stderr.write('============================================================\n')

    initTest = init_test
    preTest = pre_test
    runTest = run_test
    postTest = post_test
    cleanTest = clean_test
    startStep = start_step

    assertEqual = assert_equal
    assertMatch = assert_match
    waitForEqual = wait_for_equal
    waitForMatch = wait_for_match

    TestClassName = test_class_name
    TestName = test_name
    TestDoc = test_doc

    logInfo = log_info


class ITestCaseRunner(object):
    '''测试用例执行器接口定义
    '''

    def run(self, testcase, testresult_factory):
        '''执行一个测试用例

        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory

        :return TestResult/TestResultCollection - 测试结果
        '''
        raise NotImplementedError()


class TestCaseRunner(ITestCaseRunner):
    '''负责执行一个测试用例

    如果一个测试用例没有指定case_runner类变量，则默认都使用TestCaseRunner来执行这个用例。
    测试用例可以自定义和TestCaseRunner接口兼容的runner类，并设置case_runner类变量来实现
    自定义一个测试用例的执行逻辑，以下是TestCaseRunner的接口定义
    '''

    CLEANUP_TIMEOUT = 300

    def __init__(self):
        '''构造函数
        '''
        self._lock = threading.Lock()
        self._stop_run = False
        self._error = None

    def _rewrite_assert(self, cls):
        if not settings.get("QTAF_REWRITE_ASSERT", True) or cls == TestCase:
            return
        rewriter = AssertionRewriter()
        for key in dir(cls):
            item = getattr(cls, key)
            if isinstance(item, (types.MethodType, types.FunctionType)):
                rewriter.rewrite(item)

    def _walk_bases(self, test_case):
        '''遍历所有基类，进行相应的处理
        '''
        cur_cls = type(test_case)
        self._single_methods_mapping(cur_cls)
        self._rewrite_assert(cur_cls)
        bases = set(cur_cls.__bases__)
        while bases:
            new_bases = set()
            for base_cls in bases:
                self._single_methods_mapping(base_cls)
                self._rewrite_assert(cur_cls)
                new_bases = new_bases.union(set(base_cls.__bases__))
            bases = new_bases

    def _single_methods_mapping(self, cls_type):
        '''针对单个类进行方法映射
        '''
        if object == cls_type:
            return
        case_mothods = [
            ("initTest", "init_test"),
            ("preTest", "pre_test"),
            ("runTest", "run_test"),
            ("postTest", "post_test"),
            ("cleanTest", "clean_test")
        ]
        cls_dict = cls_type.__dict__
        for pairs in case_mothods:
            if pairs[0] in cls_dict and not pairs[1] in cls_dict:
                setattr(cls_type, pairs[1], cls_dict[pairs[0]])
                self._testresult.warning("严重警告：类型%s应当定义方法%s，方法%s已被废弃，请勿使用" % (cls_type.__name__, pairs[1], pairs[0]))
            elif pairs[0] not in cls_dict and pairs[1] in cls_dict:
                setattr(cls_type, pairs[0], cls_dict[pairs[1]])
            elif pairs[0] in cls_dict and pairs[1] in cls_dict:
                if cls_dict[pairs[0]] != cls_dict[pairs[1]]:
                    raise RuntimeError('类型%s不允许同时独立定义"%s"和"%s"' % (cls_type, pairs[0], pairs[1]))

    def _check_testcase(self, testcase):
        '''检查测试用例

        :param testcase: 测试用例
        :type testcase: TestCase
        :returns boolean, string - 检测是否通过，错误消息
        '''
        attrnames = ['owner', 'priority', 'status', 'timeout']
        tcattrs = {}
        for attr in attrnames:
            attrvalue = getattr(self._testcase, attr)
            if isinstance(attrvalue, six.text_type):
                if attrvalue.strip(" ") == "":
                    attrvalue = None
            tcattrs[attr] = attrvalue
        tcattrs['testname'] = self._testcase.test_name
        if None in tcattrs.values():
            raise RuntimeError("测试用例需要设置以下类属性：%s" % attrnames)
        if not self._testcase.test_doc:
            raise RuntimeError("测试用例的描述不能为空，请加上描述！")

        self._walk_bases(testcase)

    def _thread_run(self):
        '''测试用例线程过程
        '''
        #                     函数时发生了死锁，故注释掉。观察一段时间，看修改是否会影响测试。
        try:
            try:
                self._check_testcase(self._testcase)
            except RuntimeError as e:
                self._testresult.error(e.args[0])
                return

            while True:
                with self._lock:
                    if len(self._subtasks) == 0 or self._stop_run:
                        break
                    it = self._subtasks.popleft()

                if isinstance(it, str):
                    try:
                        if it in ['init_test', 'initTest']:
                            getattr(self._testcase, it)(self._testresult)
                        else:
                            task_result = getattr(self._testcase, it)()
                            if task_result and isinstance(task_result, six.string_types):
                                valid_task_result = False
                                for name in TestResultType.__dict__:
                                    value = TestResultType.__dict__[name]
                                    if task_result == value:
                                        valid_task_result = True
                                        break
                                if not valid_task_result:
                                    self._testresult.warning('指定的自定义状态不支持，用例将继续执行')
                                else:
                                    self._testresult.customize_result(task_result)
                                    while self._subtasks[0] not in ['post_test', 'postTest']:
                                        self._subtasks.popleft()
                    except:
                        self._testresult.exception('%s执行失败' % it)
                        if settings.get("QTAF_FAILED_SKIP_RUNTEST", False) and it in ['pre_test', 'preTest']:
                            while self._subtasks[0] not in ['post_test', 'postTest']:
                                self._subtasks.popleft()
                else:
                    it()

        except:
            self._error = traceback.format_exc()

    def _thread_cleanup(self):
        '''清理线程
        '''
        try:
            while True:
                with self._lock:
                    if len(self._subtasks) == 0:
                        break
                    it = self._subtasks.popleft()
                    if it in ['initTest', 'preTest', 'runTest', 'init_test', 'pre_test', 'run_test']:
                        continue

                if isinstance(it, str):
                    try:
                        getattr(self._testcase, it)()
                    except:
                        self._testresult.exception('用例超时时%s执行失败' % it)
                else:
                    it()
        except:
            self._error = traceback.format_exc()

    def setup(self, testcase, testresult):
        '''测试执行初始化

        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult: 测试用例结果
        :type testresult: TestResult
        '''
        pass

    def teardown(self, testcase, testresult):
        '''测试执行清理

        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult: 测试用例结果
        :type testresult: TestResult
        '''
        pass

    def run(self, testcase, testresult_factory):
        '''执行一个测试用例

        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory
        :rtype: TestResult/TestResultCollection - 测试结果
        '''
        #                       临时方案是disable gc后kill掉程序，在enable gc。后续这里
        #                        需要重新考虑已独立进程来执行测试用例。

        self._stop_run = False
        self._testcase = testcase
        self._testresult = testresult_factory.create(testcase)
        self._subtasks = collections.deque(['init_test', 'pre_test', 'run_test', 'post_test', 'clean_test'])

        with ThreadGroupScope('%s:%s' % (self._testcase.test_name, id(self))):

            ThreadGroupLocal().testcase = self._testcase
            ThreadGroupLocal().testresult = self._testresult

            self._testresult.begin_test(self._testcase)

            self.setup(self._testcase, self._testresult)

            if isinstance(self._testcase.timeout, int) or isinstance(self._testcase.timeout, float):
                timeout = self._testcase.timeout * 60
            else:
                timeout = 60

            test_thread = threading.Thread(target=self._thread_run)
            test_thread.daemon = True
            test_thread.start()
            test_thread.join(timeout)
            if test_thread.is_alive():
                self._stop_run = True
                try:
                    thread_traceback = self._get_current_traceback(test_thread)
                except:
                    self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时')
                else:
                    self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时，抓取测试线程当前堆栈',
                                                dict(traceback=thread_traceback))

                # 启动线程执行可能未执行的postTest和cleanTest
                cleanup_thread = threading.Thread(target=self._thread_cleanup)
                cleanup_thread.daemon = True
                cleanup_thread.start()
                cleanup_thread.join(self.CLEANUP_TIMEOUT)
                if cleanup_thread.is_alive():
                    try:
                        thread_traceback = self._get_current_traceback(cleanup_thread)
                    except:
                        self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时时清理超时')
                    else:
                        self._testresult.log_record(EnumLogLevel.TESTTIMEOUT, '测试用例执行超时时清理超时，抓取清理线程当前堆栈',
                                                    dict(traceback=thread_traceback))
                else:
                    if self._error:
                        raise RuntimeError("用例执行线程异常：\n%s" % smart_text(self._error))
            else:
                if self._error:
                    raise RuntimeError("用例执行线程异常：\n%s" % smart_text(self._error))

            self.teardown(self._testcase, self._testresult)

            self._testresult.end_test()

        # gc.enable()
        # gc.collect()
        return self._testresult

    def _get_current_traceback(self, thread):
        '''获取用例线程的当前的堆栈

        :param thread: 要获取堆栈的线程
        :type thread: Thread
        '''
        for thread_id, stack in sys._current_frames().items():
            if thread_id != thread.ident:
                continue
            tb = "Traceback ( thread-%d possibly hold at ):\n" % thread_id
            for filename, lineno, name, line in traceback.extract_stack(stack):
                tb += '  File: "%s", line %d, in %s\n' % (filename, lineno, name)
                if line:
                    tb += "    %s\n" % (line.strip())
            return tb
        else:
            raise RuntimeError("thread not found")


class RepeatTestCaseRunner(ITestCaseRunner):
    '''重复执行的用例执行器

    可以通过设置测试用例的类属性为此runner实例来实现指定测试用例
    执行多次。测试用例执行时可以访问成员变量iteration来判断当前是
    第几次执行。

    使用示例如下::

        class HelloRepeatTest(TestCase):
            '示例用例'
            case_runner = RepeatTestCaseRunner()
            owner = "foo"
            timeout = 1
            status = TestCase.EnumStatus.Ready
            priority = TestCase.EnumPriority.Normal

            def run_test(self):
                self.log_info("第%s次执行测试"%self.iteration)

    '''

    def __init__(self, case_runner_class=None):
        '''指定执行一个用例的CaseRunner类，不指定则为TestCaseRunner类
        '''
        if case_runner_class is None:
            case_runner_class = TestCaseRunner
        self._case_runner_class = case_runner_class

    def run(self, testcase, testresult_factory):
        '''执行一个测试用例

        :param testcase: 执行的测试用例
        :type testcase: TestCase
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory

        :returns : 测试结果
        :rtype TestResult/TestResultCollection - 测试结果
        '''
        passed = True
        if not hasattr(testcase, "repeat"):
            passed = False
            err_msg = "使用RepeatTestCaseRunner的测试用例需要设置类属性：repeat"
        elif testcase.repeat <= 0:
            passed = False
            err_msg = "使用RepeatTestCaseRunner的测试用例的类属性'repeat'必须大于0"
        if not passed:
            with ThreadGroupScope('%s:%s' % (testcase.test_name, id(self))):
                result = testresult_factory.create(testcase)
                ThreadGroupLocal().testcase = testcase
                ThreadGroupLocal().testresult = result
                result.begin_test(testcase)
                result.error(err_msg)
                result.end_test()
                return result

        passed = True
        results = []
        for i in range(testcase.repeat):
            testcase.iteration = i
            case_result = self._case_runner_class().run(testcase, testresult_factory)
            passed &= case_result.passed
            results.append(case_result)
            if not passed:
                break
        return TestResultCollection(results, passed)


class SeqTestCaseRunner(ITestCaseRunner):
    '''顺序执行的用例的执行器
    '''

    def run(self, testsuite, testresult_factory):
        '''执行一个顺序执行的测试用例套

        :param testsuite: 执行的测试用例套
        :type testsuite: SeqTestSuite
        :param testresult_factory: 测试结果工厂
        :type testresult_factory: ITestResultFactory

        :return TestResult/TestResultCollection - 测试结果
        '''
        passed = True
        results = []
        for it in testsuite:
            runner = getattr(it, 'case_runner', TestCaseRunner())
            case_result = runner.run(it, testresult_factory)
            passed &= case_result.passed
            results.append(case_result)
            if not passed:
                break
        return TestResultCollection(results, passed)


class TestSuite(object):
    '''测试用例套
    '''

    @property
    def suite_class_name(self):
        '''测试套类名称
        '''
        cls = type(self)
        if cls.__module__ == '__main__':
            type_name = cls.__name__
        else:
            type_name = (cls.__module__ + '.' + cls.__name__)
        return type_name

    def dumps(self):
        '''序列化
        '''
        raise NotImplementedError()

    def loads(self, buf):
        '''反序列化
        '''
        raise NotImplementedError()


class SeqTestSuite(TestSuite):
    '''顺序执行的测试用例套
    '''
    case_runner = SeqTestCaseRunner()

    def __init__(self, testcases):
        '''构造函数

        :param testcases: 测试用例列表
        :type testcases: list
        :param name: 测试用例名
        :type name: string
        '''
        self._testcases = testcases
        self._resmgr = None
        self.__share_data_mgr = None

    def __iter__(self):
        for it in self._testcases:
            yield it

    def __len__(self):
        return len(self._testcases)

    def __repr__(self):
        return '<SeqTestSuite module:%s>' % self.test_class_name

    @property
    def test_class_name(self):
        '''返回测试用例名字（不同测试用例的名字不同）

        :rtype: str
        '''
        cls = type(self._testcases[0])
        return cls.__module__

    @property
    def test_name(self):
        '''返回测试用例实例的名字

        :rtype: str
        '''
        cls = type(self._testcases[0])
        return cls.__module__

    @property
    def test_doc(self):
        '''测试用例说明

        :rtype: str
        '''
        cls = type(self._testcases[0])
        desc = cls.__module__.__doc__
        if isinstance(desc, six.text_type):
            desc = re.sub('^\s*', '', desc)
            desc = re.sub('\s*$', '', desc)
        return desc

    @property
    def test_result(self):
        '''将最后一个执行的用例结果，作为Suite的结果
        '''
        result = None
        for testcae in self._testcases:
            if testcae.test_result:
                result = testcae.test_result
            else:
                break
        return result

    @property
    def test_resmgr(self):
        '''资源管理器
        '''
        return self._resmgr

    @test_resmgr.setter
    def test_resmgr(self, resmgr):
        self._resmgr = resmgr
        for it in self._testcases:
            it.test_resmgr = resmgr


    @property
    def share_data_mgr(self):
        '''共享数据管理器
        '''
        return self.__share_data_mgr

    @share_data_mgr.setter
    def share_data_mgr(self, share_data_mgr):
        self.__share_data_mgr = share_data_mgr
        for it in self._testcases:
            it._share_data_mgr = share_data_mgr

    def dumps(self):
        '''序列化
        '''
        from testbase import serialization
        return [serialization.dumps(it) for it in self._testcases]

    def loads(self, buf):
        '''反序列化
        '''
        from testbase import serialization
        self._testcases = [serialization.loads(it) for it in buf]


def debug_run_all():
    '''调试执行当前脚本的全部用例
    '''
    from testbase.loader import TestLoader
    from testbase.runner import TestRunner
    from testbase.report import StreamTestReport
    tests = TestLoader().load("__main__")
    runner = TestRunner(StreamTestReport(output_testresult=True, output_summary=True))
    runner.run(tests)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
共用类模块
'''

import binascii
import codecs
import importlib
import inspect
import io
import locale
import os
import pkg_resources
import re
import six
import sys
import threading
import time
import traceback

from inspect import isclass, getmro, types
from xml.dom.minidom import Node
from datetime import datetime

from tuia.exceptions import TimeoutError

default_locale = locale.getdefaultlocale()
if default_locale:
    default_encoding = default_locale[1] or "utf-8"
else:
    default_encoding = "utf-8"

file_encoding = sys.getfilesystemencoding()
file_encoding_lower = file_encoding.lower()
for special_encoding in ["ansi", "ascii"]:
    if file_encoding_lower.find(special_encoding) >= 0:
        file_encoding = "utf-8"
        break
file_encoding = file_encoding or "utf-8"


class Timeout(object):
    '''TimeOut类，实现超时重试逻辑
    '''

    def __init__(self, timeout=10, interval=0.5):
        '''Constructor

        :param timeout: 超时秒数，默认是10
        :param interval: 重试时间间隔秒数，默认是0.5
        '''
        self.timeout = float(timeout)
        self.interval = float(interval)

    def retry(self,
        func,
        args,
        exceptions=(),
        resultmatcher=None,
        nothrow=False):
        """多次尝试调用函数，成功则并返回调用结果，超时则根据选项决定抛出TimeOutError异常。

        :param func: 尝试调用的函数
        :type args: dict或tuple
        :param args: func函数的参数
        :type exceptions: tuple类型，tuple元素是异常类定义，如QPathError, 而不是异常实例，如QPathError()
        :param exceptions: 调用func时抛出这些异常，则重试。
                                                                        如果是空列表(),则不捕获异常。
        :type resultmatcher: 函数指针类型
        :param resultmatcher: 函数指针，用于验证第1个参数func的返回值。
                                                                                默认值为None，表示不验证func的返回值，直接返回。
                                                                               其函数原型为:
                                  def result_match(ret):  # 参数ret为func的返回值
                                      pass
                                                                               当result_match返回True时，直接返回，否则继续retry。
        :type  nothrow:bool
        :param nothrow:如果为True，则不抛出TimeOutError异常

        :return: 返回成功调用func的结果
        :rtype : any
        """
        start = time.time()
        waited = 0.0
        try_count = 0
        ret = None
        while True:
            try:
                try_count += 1
                if dict == type(args):
                    ret = func(**args)
                elif tuple == type(args):
                    ret = func(*args)
                else:
                    raise TypeError("args type %s is not a dict or tuple" % type(args))

                if resultmatcher == None or resultmatcher(ret) == True:
                    return ret
            except exceptions:
                pass

            waited = time.time() - start
            if waited < self.timeout:
                time.sleep(min(self.interval, self.timeout - waited))
            elif try_count == 1 :
                continue
            else:
                if nothrow:
                    return ret
                else:
                    raise TimeoutError("在%d秒里尝试了%d次" % (self.timeout, try_count))

    def waitObjectProperty(self, obj, property_name, waited_value, regularMatch=False):
        '''通过比较obj.property_name和waited_value，等待属性值出现。
                             如果属性值obj.property_name是字符类型则waited_value做为正则表达式进行比较。
                             比较成功则返回，超时则抛出TimeoutError异常。

        :param obj: 对象
        :param property_name: 要等待的obj对象的属性名
        :param waited_value: 要比较的的属性值，支持多层属性
        :param regularMatch: 参数 property_name和waited_value是否采用正则表达式的比较。
                                                                            默认为不采用（False）正则，而是采用恒等比较
        '''
        start = time.time()
        waited = 0.0
        try_count = 0
        isstr = isinstance(waited_value, six.string_types)
        while True:
            objtmp = obj  # 增加多层属性支持
            pro_names = property_name.split('.')
            for i in range(len(pro_names)):
                propvalue = getattr(objtmp, pro_names[i])
                objtmp = propvalue

            if isstr and regularMatch:  # 简化原逻辑
                if None != re.search(waited_value, propvalue):
                    return
            else:
                if waited_value == propvalue:
                    return
            try_count += 1
            waited = time.time() - start
            if waited < self.timeout:
                time.sleep(min(self.interval, self.timeout - waited))
            else:
                raise TimeoutError("对象属性值比较超时（%d秒%d次）：期望值:%s，实际值:%s，"
                                   % (self.timeout, try_count, waited_value, propvalue))

    def check(self, func, expect):
        '''多次检查func的返回值是否符合expect设定的期望值，如果设定时间内满足，则返回True，否则返回False

        :param func: 尝试调用的函数
        :param expect: 设定的期望值

        :returns bool - 检查是否符合预期
        '''
        start = time.time()
        waited = 0.0
        while True:
            if expect == func():
                return True
            waited = time.time() - start
            if waited < self.timeout:
                time.sleep(min(self.interval, self.timeout - waited))
            else:
                return False


class Singleton(type):
    """单实例元类，用于某个类需要实现单例模式。
            使用方式示例如下::
          import six
          class MyClass(with_metaclass(Singleton, object)):
              def __init__(self, *args, **kwargs):
                  pass

    """
    _instances = {}

    def __init__(cls, name, bases, dic):
        super(Singleton, cls).__init__(name, bases, dic)
        cls._instances = {}

    def __call__(self, *args, **kwargs):
        if self not in self._instances:
            self._instances[self] = super(Singleton, self).__call__(*args, **kwargs)
        return self._instances[self]


class LazyInit(object):
    '''实现延迟初始化

    使用方式示例::

        class _Win32Window(object)
            def click(self):
                #......
        class Control(object):
            def __init__(self, locator):
                self._locator = locator
                self._initobj = LazyInit(self, '_initobj', self._init_window)
            def _init_window(self):
                return _Win32Window(self._locator)
            def click(self):
                return self._initobj.click()

        ctrl = Control("/Name=xxx)
        ctrl.click()  # <-- call _init_window
        ctrl.click()

    '''

    def __init__(self, obj, propname, init_func):
        '''构造函数
        '''
        self.__obj = obj
        self.__propname = propname
        self.__init_func = init_func

    def __getattr__(self, attrname):
        obj = self.__init_func()
        setattr(self.__obj, self.__propname, obj)
        attr = getattr(obj, attrname)
        del self.__obj
        del self.__propname
        del self.__init_func
        return attr

    def __setattr__(self, attrname, value):
        if attrname in ['_LazyInit__obj', '_LazyInit__propname', '_LazyInit__init_func']:
            return super(LazyInit, self).__setattr__(attrname, value)
        obj = self.__init_func()
        setattr(self.__obj, self.__propname, obj)
        setattr(obj, attrname, value)
        del self.__obj
        del self.__propname
        del self.__init_func


class ShareDataManager(object):
    def __init__(self, lock=threading.Lock(), data={}):
        self._data = data
        self._lock = lock

    @property
    def data(self):
        return self._data

    def get(self, key):
        self._lock.acquire()
        data = self._data.get(key)
        self._lock.release()
        if not data:
            raise KeyError("No such key %s exists" % key)
        return data.get('value', None)

    def set(self, key, value, level=0):
        self._lock.acquire()
        self._data[key] = {
            "value": value,
            "level": level
        }
        self._lock.release()

    def pop(self, key):
        try:
            self._lock.acquire()
            data = self._data.pop(key)
        except KeyError:
            raise KeyError("No such key %s exists" % key)
        finally:
            self._lock.release()
        return data.get("value", None)

    def remove(self, key):
        try:
            self._lock.acquire()
            self._data.pop(key)
        except KeyError:
            raise KeyError("No such key %s exists" % key)
        finally:
            self._lock.release()


class ThreadGroupLocal(object):
    '''使用线程组本地存储的元类

    - 当配合ThreadGroupScope使用，类似threading.local()提供的TLS变种，一个线程和其子孙线程共享一个存储
    详细使用方式请参考ThreadGroupScope类

    - 当不在ThreadGroupScope中使用时，行为和threading.local()一致

    '''

    def __init__(self):
        curr_thread = threading.current_thread()
#         if not hasattr(curr_thread, 'qtaf_group'):
#             raise RuntimeError("current thread is not in any QTAF thread group scope")
#         self.__data = curr_thread.qtaf_local
        if hasattr(curr_thread, 'qtaf_group'):
            self.__data = curr_thread.qtaf_local
        else:
            if not hasattr(curr_thread, 'qtaf_local_outofscope'):
                curr_thread.qtaf_local_outofscope = {}
            self.__data = curr_thread.qtaf_local_outofscope

    def __setattr__(self, name, value):
        if name.startswith('_ThreadGroupLocal__'):
            super(ThreadGroupLocal, self).__setattr__(name, value)
        else:
            self.__data[name] = value

    def __getattr__(self, name):
        if name.startswith('_ThreadGroupLocal__'):
            return super(ThreadGroupLocal, self).__getattr__(name)
        else:
            try:
                return self.__data[name]
            except KeyError:
                raise AttributeError("'ThreadGroupLocal' object has no attribute '%s'" % (name))


class ThreadGroupScope(object):
    '''指定线程组作用域，进入这个作用域的线程，以及在其作用域内创建的线程都同属于一个线程组

    使用示例如下::

        def _thread_proc():
            ThreadGroupLocal().counter +=1

        with ThreadGroupScope("test_group"):
            ThreadGroupLocal().counter = 0
            t = threading.Thread(target=_thread_proc)
            t.start()
            t.join()
            t = threading.Thread(target=_thread_proc)
            t.start()
            t.join()
            assert ThreadGroupLocal().counter == 2

    '''

    def __init__(self, name):
        '''构造函数

        :param name: 线程组名称，全局唯一
        :type name: string
        '''
        self._name = name

    def __enter__(self):
        curr_thread = threading.current_thread()
        if hasattr(curr_thread, 'qtaf_local'):
            raise RuntimeError("ThreadGroupScope cannot be nested")
        curr_thread.qtaf_local = {}
        curr_thread.qtaf_group = self._name

    def __exit__(self, *exc_info):
        del threading.current_thread().qtaf_local
        del threading.current_thread().qtaf_group

    @staticmethod
    def current_scope():
        '''返回当前线程所在的线程组作用域，如果不存在于任务线程组作用域，则返回None
        '''
        curr_thread = threading.current_thread()
        if hasattr(curr_thread, 'qtaf_group'):
            return curr_thread.qtaf_group


_origin_thread_start_func = threading.Thread.start


def _thread_start_func(self, *args, **kwargs):
    '''用于劫持threading.Thread.start函数
    '''
    curr_thread = threading.current_thread()
    if hasattr(curr_thread, 'qtaf_group'):
        self.qtaf_group = curr_thread.qtaf_group
        self.qtaf_local = curr_thread.qtaf_local
    return _origin_thread_start_func(self, *args, **kwargs)


threading.Thread.start = _thread_start_func


def ForbidOverloadMethods(func_name_list):
    '''生成metaclass用于指定基类禁止子类重载函数
    '''

    class _metaclass(type):

        def __init__(cls, name, bases, dic):
            if len(bases) == 1 and bases[0] == object:
                super(_metaclass, cls).__init__(name, bases, dic)
            else:
                for it in func_name_list:
                    if it in dic.keys():
                        raise RuntimeError("不允许%s重载函数: %s" % (cls.__name__, it))
                super(_metaclass, cls).__init__(name, bases, dic)

    return _metaclass


class classproperty(object):
    '''类属性修饰器
    '''

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def smart_text(s, decoding=None):
    '''convert any text or binary to text
    py2 text: utf-8 bytes
    py3 text: unicode
    '''
    if not isinstance(s, (six.string_types, six.binary_type)):
        raise RuntimeError("string or binary type didn't match with %r" % s)
    if six.PY3:
        if isinstance(s, six.text_type):
            return s
        else:
            try:
                return s.decode('utf8')
            except UnicodeDecodeError:  # other encoding
                try:
                    if decoding is None:
                        decoding = default_encoding
                    return s.decode(decoding)
                except UnicodeDecodeError:  # mixed encoding
                    return repr(s)
    else:
        return smart_binary(s, decoding=decoding)  # py2


def  smart_binary(s, encoding="utf8", decoding=None):
    '''convert any text or binary to binary of specified encoding
    '''
    if not isinstance(s, (six.string_types, six.binary_type)):
        raise RuntimeError("string or binary type didn't match with %r" % s)
    if isinstance(s, six.text_type):
        try:
            return s.encode(encoding)
        except UnicodeEncodeError:
            if six.PY3:
                return bytes(repr(s), encoding)
            else:
                return bytes(repr(s))

    assert type(s) == six.binary_type
    try:
        return s.decode("utf8").encode(encoding)  # other encoding
    except UnicodeError:
        try:
            if decoding is None:
                decoding = default_encoding
            return s.decode(decoding).encode(encoding)
        except UnicodeError:  # data mixed encoding
            if six.PY3:
                return bytes(repr(s), encoding)
            else:
                return bytes(repr(s))


def smart_bytes(data):
    if not isinstance(data, (six.string_types, six.binary_type)):
        raise ValueError("data=%r does not match string or bytes type" % data)
    if isinstance(data, six.text_type):
        return smart_binary(data)
    else:
        return data


def smart_to_hex(s, encoding="utf8", decoding=None):
    s = smart_binary(s, encoding=encoding, decoding=decoding)
    binary_s = binascii.hexlify(s)
    return smart_text(binary_s)


def smart_from_hex(s):
    s = smart_binary(s)
    binary_s = binascii.unhexlify(s)
    return smart_text(binary_s)


def smart_bytify(obj, encoding="utf-8", decoding=None):
    """recursively convert objects from string types to binary
    """
    if isinstance(obj, dict):
        dic = {}
        for key, value in obj.items():
            dic[smart_bytify(key, encoding, decoding)] = smart_bytify(value, encoding, decoding)
        return dic
    elif isinstance(obj, list):
        ls = []
        for element in obj:
            ls.append(smart_bytify(element, encoding, decoding))
        return ls
    elif isinstance(obj, six.string_types):
        return smart_binary(obj, encoding, decoding)
    else:
        return obj


def smart_strfy(obj, decoding=None):
    """recursively convert objects from binary to text
    """
    if isinstance(obj, dict):
        dic = {}
        for key, value in obj.items():
            dic[smart_strfy(key, decoding)] = smart_strfy(value, decoding)
        return dic
    elif isinstance(obj, list):
        ls = []
        for element in obj:
            ls.append(smart_strfy(element, decoding))
        return ls
    elif isinstance(obj, six.string_types):
        return smart_text(obj, decoding)
    else:
        return obj


def get_thread_traceback(thread):
    '''获取用例线程的当前的堆栈

    :param thread: 要获取堆栈的线程
    :type thread: Thread
    '''
    for thread_id, stack in sys._current_frames().items():
        if thread_id != thread.ident:
            continue
        tb = "Traceback ( thread-%d possibly hold at ):\n" % thread_id
        for filename, lineno, name, line in traceback.extract_stack(stack):
            tb += '  File: "%s", line %d, in %s\n' % (filename, lineno, name)
            if line:
                tb += "    %s\n" % (line.strip())
        return tb
    else:
        raise RuntimeError("thread not found")


def get_method_defined_class(method):
    method_name = method.__name__
    need_prefix = False
    if method_name.startswith("__"):
        if method_name != "__init__":
            need_prefix = True
    if getattr(method, "__self__", None):
        classes = [method.__self__.__class__]
    else:
        # unbound method
        if hasattr(method, "im_class"):
            classes = [method.im_class]
        else:
            classes = [getattr(method, "__class__", object)]
    while classes:
        c = classes.pop()
        if c == object:
            if len(classes) > 0:
                continue
        if need_prefix:
            target_name = "_" + c.__name__ + method_name
        else:
            target_name = method_name
        if target_name in c.__dict__:
            return c
        else:
            classes = list(c.__bases__) + classes
    raise RuntimeError("cannot find implement class for method: %s" % method)


def get_last_frame_stack(back_count=2):
    frame = inspect.currentframe()
    for _ in range(back_count):
        frame = frame.f_back
    stack = "".join(traceback.format_stack(frame, 1))
    return stack


def to_pretty_xml(doc, encoding="utf-8"):
    """we need to ensure each line to be binary type
    """

    class _XMLWriter(codecs.StreamWriter):
        """an inner writer to give writer a chance to handle each line
        """

        def write(self, data):
            data = smart_binary(data)
            self.stream.write(data)

    buff = io.BytesIO()
    indent = "    "
    newl = "\n"
    writer = _XMLWriter(buff)
    if doc.nodeType == Node.DOCUMENT_NODE:  # document node needs encoding
        doc.writexml(writer, "", indent, newl, encoding)
    else:
        doc.writexml(writer, "", indent, newl)
    return writer.stream.getvalue()


def ensure_binary_stream(stream, encoding="utf-8"):
    if six.PY3:
        stream_type = type(stream)
        if issubclass(stream_type, io.IOBase):
            if not hasattr(stream, "mode"):
                raise ValueError("stream=%r does not have attribute mode" % stream)
            if "b" in stream.mode:
                new_stream = stream
            else:
                new_stream = stream.buffer
        else:  # making our effort to adapt to the abnormal stream
            orig_stream_write_func = stream.write

            def _binary_write(s):
                s = smart_text(s)
                if not isinstance(s, bytes):
                    s = s.encode(encoding)
                orig_stream_write_func(s)

            stream.write = _binary_write
            new_stream = stream
    else:
        if getattr(stream, "encoding", None):
            if not stream.encoding.lower().startswith('ansi'):  # linux ascii
                encoding = stream.encoding
        new_stream = stream
    return new_stream, encoding


def codecs_open(filename, mode="rb", encoding=None, errors="strict", buffering=1):
    filename = smart_binary(filename, encoding=file_encoding)
    return codecs.open(filename, mode=mode, encoding=encoding, errors=errors, buffering=buffering)


def path_exists(filename):
    filename = smart_binary(filename, encoding=file_encoding)
    return os.path.exists(filename)


def get_os_version():
    if sys.platform == 'win32':
        with os.popen("ver") as pipe:
            osver = smart_text(pipe.read())
    else:
        osver = smart_text(str(os.uname()))  # @UndefinedVariable
    return osver


if six.PY3:
    maketrans_func = str.maketrans
else:
    import string
    maketrans_func = string.maketrans

BAD_FILE_CHARS = r'\/*?:<>"|~#'
BAD_VAR_CHAR = BAD_FILE_CHARS + '()[]+-=& '
TRANS = maketrans_func(BAD_VAR_CHAR, '_' * len(BAD_VAR_CHAR))
BAD_VAR_CHAR_SET = set(BAD_VAR_CHAR)


def translate_bad_char(input_string):
    if six.PY2:
        translated_string = smart_binary(input_string).translate(TRANS)
        translated_string = re.sub(r'[^\x00-\x7f]', '', translated_string) # Replace non-ascii chars
    else:
        translated_string = smart_text(input_string).translate(TRANS)
    return translated_string


def has_bad_char(input_string):
    if set(input_string) & BAD_VAR_CHAR_SET:
        return True
    else:
        return False


def get_time_str():
    now = datetime.now()
    time_str = now.strftime("%Y%m%d_%H%M%S%f")[:-3]
    return time_str


def get_inner_resource(resource_module, resource_name):
    if not os.path.isfile(__file__):  # using egg
        file_path = pkg_resources.resource_filename(resource_module, resource_name)
    else:
        mod = importlib.import_module(resource_module)
        mod_path = mod.__path__[0]
        file_path = os.path.join(mod_path, resource_name)
    if not path_exists(file_path):
        raise RuntimeError("resource path: %s not existed." % file_path)
    return file_path


def get_attribute_from_string(object_path):
    parts = object_path.split(".")
    parts_len = len(parts)
    mod = None
    for index in range(parts_len):
        mod_path = ".".join(parts[:index + 1])
        try:
            mod = importlib.import_module(mod_path)
        except ImportError:
            break
    value = mod
    for new_index in range(index, parts_len):
        try:
            value = getattr(value, parts[new_index])
        except AttributeError:
            raise AttributeError("%s has no attribute or submodule named \"%s\"" % (value, parts[new_index]))
    return value


def getmembers(object, predicate=None):
    """Return all members of an object as (name, value) pairs sorted by name.
    Optionally, only return members that satisfy a given predicate."""
    if isclass(object):
        mro = (object,) + getmro(object)
    else:
        mro = ()
    results = []
    processed = set()
    names = dir(object)
    # :dd any DynamicClassAttributes to the list of names if object is a class;
    # this may result in duplicate entries if, for example, a virtual
    # attribute with the same name as a DynamicClassAttribute exists
    try:
        for base in object.__bases__:
            for k, v in base.__dict__.items():
                if isinstance(v, types.DynamicClassAttribute):
                    names.append(k)
    except:
        pass
    for key in names:
        # First try to get the value via getattr.  Some descriptors don't
        # like calling their __get__ (see bug #1785), so fall back to
        # looking in the __dict__.
        try:
            value = getattr(object, key)
            # handle the duplicate key
            if key in processed:
                raise AttributeError
        except AttributeError:
            for base in mro:
                if key in base.__dict__:
                    value = base.__dict__[key]
                    break
            else:
                # could be a (currently) missing slot member, or a buggy
                # __dir__; discard and move on
                continue
        except:
            continue
        if not predicate or predicate(value):
            results.append((key, value))
        processed.add(key)
    results.sort(key=lambda pair: pair[0])
    return results


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""重试机制实现
"""

import time
    
class _RetryItem(object):
    """
    """
    def __init__(self, iteration, timestamps=None):
        self.__iteration = iteration
        self.__ts = timestamps

    @property
    def iteration(self):
        return self.__iteration
    
    @property
    def ts(self):
        return self.__ts
    
    def __str__(self):
        return "<%s iter=%s, ts=%s>" % (self.__class__.__name__, self.__iteration, self.__ts)
            
    
class _RetryWithTimeout(object):
    """retry mechanism with timeout
    """
    def __init__(self,interval=5, timeout=60, raise_error=True):
        self.interval = interval
        self.timeout = timeout
        self.raise_error = raise_error
        self.__start_time = None
        self.__count = 0
    
    def __iter__(self):
        return self
    
    def next(self):
        if self.__start_time == None:
            self.__count += 1
            self.__start_time = time.time()
            return _RetryItem(self.__count, self.__start_time)
        else:
            time.sleep(self.interval)
            ts = time.time()
            if ts - self.__start_time < self.timeout:
                self.__count += 1
                return _RetryItem(self.__count, ts)
            else:
                if self.raise_error:
                    raise RetryLimitExcceeded("Procedure retried %s times in %ss" % (self.__count, self.timeout))
                else:
                    raise StopIteration
                
    __next__ = next
        
        
class _RetryWithCount(object):
    """retry mechanism with count
    """
    def __init__(self, limit=3, interval=0,raise_error=True):
        self.limit = limit
        self.raise_error = raise_error
        self.interval = interval
        self.__count = 0
        
    def __iter__(self):
        return self
    
    def next(self):
        self.__count += 1
        if self.__count <= self.limit:
            if self.__count !=1 and self.interval:
                time.sleep(self.interval)
            return _RetryItem(self.__count, time.time())
        if self.raise_error:
            raise RetryLimitExcceeded("Procedure retried for %s times with interval=%ss" % (self.limit, self.interval))
        else:
            raise StopIteration
        
    __next__ = next

        
class RetryLimitExcceeded(Exception):
    """maxmium retry limit excceeded error
    """
    pass

 
class Retry(object):
    """retry mechanism with timeout or limit
    """
    def __init__(self, timeout=10, limit=None, interval=0.5, raise_error=True):
        if limit:
            self._retry = _RetryWithCount(limit=limit, interval=interval, raise_error=raise_error)
        else:
            self._retry = _RetryWithTimeout(timeout=timeout, interval=interval, raise_error=raise_error)
        
    def __iter__(self):
        return self._retry.__iter__()
    
    def call(self, callee, *args, **kwargs):
        if not hasattr(callee, "__call__"):
            raise ValueError("callee must be a function instance")
        for _ in self:
            r = callee(*args, **kwargs)
            if r:
                return r



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
"""
管理和辅助工具
"""

import argparse
import inspect
import json
import os
import re
import traceback
import sys
import shlex
import six

from testbase import logger
from testbase import project
from testbase.conf import settings
from testbase.version import version
from testbase.exlib import ExLibManager
from testbase.dist import DistGenerator, VirtuelEnv
from testbase.loader import TestLoader
from testbase.testcase import TestCasePriority, TestCaseStatus, TestCase
from testbase.runner import TestCaseSettings
from testbase.report import test_list_types
from testbase.types import runner_types, report_types, resmgr_backend_types
from testbase.util import codecs_open, path_exists


class ArgumentParser(object):
    """参数解析
    """
    USAGE = """Usage: %(ProgramName)s subcommand [options] [args]

Options:
  -h, --help            show this help message and exit

Type '%(ProgramName)s help <subcommand>' for help on a specific subcommand.

Available subcommands:

%(SubcmdList)s

"""

    def __init__(self, subcmd_classes):
        """构造函数
        """
        self.subcmd_classes = subcmd_classes
        self.prog = os.path.basename(sys.argv[0])

    def print_help(self):
        """打印帮助文档
        """
        logger.info(self.USAGE % {"ProgramName": self.prog,
                                 "SubcmdList": "\n".join([ '\t%s' % it.name for it in self.subcmd_classes])})

    def parse_args(self, args):
        """解析参数
        """
        if len(args) < 1:
            self.print_help()
            return 1

        subcmd = args[0]
        for it in self.subcmd_classes:
            if it.name == subcmd:
                subcmd_class = it
                parser = it.parser
                break
        else:
            logger.error("invalid subcommand \"%s\"\n" % subcmd)
            return 1

        ns = parser.parse_args(args[1:])
        subcmd = subcmd_class()
        subcmd.main_parser = self
        return subcmd, ns

#     def add_subcommand(self, subcmd, parser ):
#         """增加一个子命令
#         """
#         parser.prog = "%s help" % self.prog
#         self._subcmd_parser_dict[subcmd] = parser

    def get_subcommand(self, name):
        """获取子命令
        """
        for it in self.subcmd_classes:
            if it.name == name:
                return it()


class Command(object):
    """一个命令
    """
    name = None
    parser = None

    def execute(self, args):
        """执行过程
        """
        raise NotImplementedError()


class Help(Command):
    """帮助命令
    """
    name = 'help'
    parser = argparse.ArgumentParser("Display subcommand usage")
    parser.add_argument('subcommand', nargs='?', help="target subcommand to display")

    def execute(self, args):
        """执行过程
        """
        if args.subcommand == None:
            self.main_parser.print_help()
        else:
            subcmd = self.main_parser.get_subcommand(args.subcommand)
            subcmd.parser.print_help()


class RunScript(Command):
    """执行一个脚本
    """
    name = 'runscript'
    parser = argparse.ArgumentParser("Run python script")
    parser.add_argument('script_path', help="target script to run")

    def execute(self, args):
        """执行过程
        """
        import runpy
        if not os.path.isfile(args.script_path):
            print("invalid file path: %s" % args.script_path)
            return 1
        runpy.run_path(args.script_path, run_name='__main__')


class RunTest(Command):
    """批量执行测试用例
    """
    name = 'runtest'
    parser = argparse.ArgumentParser("Run QTA testcases")
    parser.add_argument("tests", metavar="TEST", nargs='*', help="testcase set to executive, eg: zoo.xxx.HelloTest")
    parser.add_argument('-w', '--working-dir', default=None, help="working directory to store all result files", dest="working_dir")
    parser.add_argument('--priority', help="run test cases with specific priority, accept multiple options",
                        dest="priorities", action="append",
                        choices=[TestCasePriority.BVT, TestCasePriority.High, TestCasePriority.Normal, TestCasePriority.Low])
    parser.add_argument('--status', default=None, help="run test cases with specific status, accept multiple options",
                        dest="status", action="append",
                        choices=[TestCaseStatus.Design, TestCaseStatus.Implement, TestCaseStatus.Ready, TestCaseStatus.Review, TestCaseStatus.Suspend])
    parser.add_argument("--excluded-name", help="exclude test cases with specific name prefix , accept multiple options",
                        action="append", dest="excluded_names", metavar="EXCLUDED_NAME")
    parser.add_argument("--owner", help="run test cases with specific owner, accept multiple options",
                        action="append", dest="owners", metavar="OWNER")
    parser.add_argument("--tag", help="run test cases with specific tag, accept multiple options",
                        action="append", dest="tags", metavar="TAG")
    parser.add_argument("--excluded-tag", help="exclude test cases with specific name tag, accept multiple options",
                        action="append", dest="excluded_tags", metavar="EXCLUDED_TAG")

    parser.add_argument("--report-type", help="report type", choices=report_types.keys(), default="stream")
    parser.add_argument("--report-args", help="additional arguments for specific report", default="")
    parser.add_argument("--report-args-help", help="show help information for specific report arguemnts", choices=report_types.keys())

    parser.add_argument("--resmgr-backend-type", help="test resource manager backend type", choices=resmgr_backend_types.keys(), default="local")

    parser.add_argument("--runner-type", help="test runner type", choices=runner_types.keys(), default="basic")
    parser.add_argument("--runner-args", help="additional arguments for specific runner", default="")
    parser.add_argument("--runner-args-help", help="show help information for specific runner arguemnts", choices=runner_types.keys())

    parser.add_argument("--execute-type", help="execute type", choices=["random", "sequential"], default="random")

    parser.add_argument("--share-data", help="share data", default="")
    parser.add_argument("--global-parameters", help="global parameters", default="")

    parser.add_argument("--config-file", help="runtime config file path")

    def run_args_parser(self, runner_args):
        """兼容参数传入concurrency=5,retries=1，支持subprocess shell=False"""
        regex_c = re.compile(r'(\w+=\w+)+')
        regex = regex_c.search(runner_args)
        if regex:
            # concurrency=5,retries=1
            ret = []
            args = regex_c.findall(runner_args)
            for arg in args:
                tmp = arg.split("=")
                ret.append("--%s %s" % (tmp[0], tmp[1]))
            return " ".join(ret)
        else:
            # --concurrency 5 --retries 1
            return runner_args

    def execute(self, args):
        """执行过程
        """
        if args.config_file:
            with open(args.config_file, 'r') as fp:
                data = json.load(fp)
                for k, value in data.items():
                    setattr(args, k, value)

        if args.report_args_help:
            report_types[args.report_args_help].get_parser().print_help()
            return 1
        if args.runner_args_help:
            runner_types[args.runner_args_help].get_parser().print_help()
            return 1
        if not args.tests:
            logger.info("no test set specified")
            return 1

        if args.working_dir is None:
            args.working_dir = os.getcwd()
        prev_dir = os.getcwd()
        if not path_exists(args.working_dir):
            os.makedirs(args.working_dir)
        os.chdir(args.working_dir)

        priorities = args.priorities or [TestCase.EnumPriority.Low,
                                          TestCase.EnumPriority.Normal,
                                          TestCase.EnumPriority.High,
                                          TestCase.EnumPriority.BVT]

        status = args.status or [TestCase.EnumStatus.Design,
                                  TestCase.EnumStatus.Implement,
                                  TestCase.EnumStatus.Review,
                                  TestCase.EnumStatus.Ready]

        if args.share_data and isinstance(args.share_data, six.string_types):
            args.share_data = json.loads(args.share_data)

        if args.global_parameters and isinstance(args.global_parameters, six.string_types):
            args.global_parameters = json.loads(args.global_parameters)

        test_conf = TestCaseSettings(names=args.tests,
                                    excluded_names=args.excluded_names,
                                    priorities=priorities,
                                    status=status,
                                    owners=args.owners,
                                    tags=args.tags,
                                    excluded_tags=args.excluded_tags,
                                    global_parameters=args.global_parameters)

        report_type = report_types[args.report_type]
        if args.report_type == 'xml':

            class VerboseXMLTestReport(report_types[args.report_type]):

                def log_test_result(self, testcase, testresult):
                    logger.info("run test case: %s(pass?:%s)" % (testcase.test_name, testresult.passed))
                    super(VerboseXMLTestReport, self).log_test_result(testcase, testresult)

            report_type = VerboseXMLTestReport

        elif args.report_type == 'online':

            class VerboseOnlineTestReport(report_types[args.report_type]):

                def log_test_result(self, testcase, testresult):
                    logger.info("run test case: %s(pass?:%s)" % (testcase.test_name, testresult.passed))
                    super(VerboseOnlineTestReport, self).log_test_result(testcase, testresult)

                def begin_report(self):
                    super(VerboseOnlineTestReport, self).begin_report()
                    print("report url: %s" % self.url)
                    with codecs_open(os.path.join(os.getcwd(), "report_url.txt"), "w", encoding="utf-8") as fd:
                        fd.write(self.url)

            report_type = VerboseOnlineTestReport

        report = report_type.parse_args(shlex.split(args.report_args))
        resmgr_backend = resmgr_backend_types[args.resmgr_backend_type]()
        execute_type = args.execute_type

        runner_type = runner_types[args.runner_type]
        # 解析runner_args参数, 用于支持命令行传入concurrency=5,retries=1
        runner_args = self.run_args_parser(args.runner_args)
        runner = runner_type.parse_args(shlex.split(runner_args), report, resmgr_backend, execute_type)

        runner.run(test_conf)
        os.chdir(prev_dir)
        if args.report_type == 'online':
            if sys.platform == "win32" and os.environ.get("AUTO_OPEN_URL", "1") == "1":
                logger.info("opening online report url:%s" % report.url)
                os.system("start %s" % report.url)
            else:
                logger.info("online report generated: %s" % report.url)

        elif args.report_type == 'xml':
            if sys.platform == "win32" and os.environ.get("AUTO_OPEN_URL", "1") == "1":
                logger.info("opening XML report with IE...")
                report_xml = os.path.abspath(os.path.join(args.working_dir, "TestReport.xml"))
                os.system("start iexplore %s" % report_xml)
            else:
                logger.info("XML report generated: %s" % os.path.abspath("TestReport.xml"))

        if not report.is_passed():
            return 1


class DiscoverTests(Command):
    """discover tests
    """
    name = "discover"
    parser = argparse.ArgumentParser("Discover tests")
    parser.add_argument("tests", metavar="TEST", nargs='*', help="testcase set to executive, eg: zoo.xxx.HelloTest")
    parser.add_argument('--priority', help="test cases with specific priority, accept multiple options, default is High and Normal",
                        dest="priorities", action="append",
                        choices=[TestCasePriority.BVT, TestCasePriority.High, TestCasePriority.Normal, TestCasePriority.Low])
    parser.add_argument('--status', help="test cases with specific status, accept multiple options, default is Ready",
                        dest="status", action="append",
                        choices=[TestCaseStatus.Design, TestCaseStatus.Implement, TestCaseStatus.Ready, TestCaseStatus.Review, TestCaseStatus.Suspend])
    parser.add_argument("--excluded-name", help="exclude test cases with specific name prefix , accept multiple options",
                        action="append", dest="excluded_names", metavar="EXCLUDED_NAME")
    parser.add_argument("--owner", help="test cases with specific owner, accept multiple options",
                        action="append", dest="owners", metavar="OWNER")
    parser.add_argument("--tag", help="test cases with specific tag, accept multiple options",
                        action="append", dest="tags", metavar="TAG")
    parser.add_argument("--excluded-tag", help="exclude test cases with specific name tag, accept multiple options",
                        action="append", dest="excluded_tags", metavar="EXCLUDED_TAG")
    parser.add_argument("--show", help="explicitly specify test case types to show, default will show all kinds of test cases",
                        choices=["filtered", "error", "normal"], dest="shows", action="append")
    parser.add_argument("--output-file", help="output file name, default is stdout", default=None)
    parser.add_argument("--output-type", help="output type, default is stream", default="stream", choices=test_list_types.keys())

    def execute(self, args):
        if not args.tests:
            logger.info("no test set specified")
            return 1
        shows = args.shows or ["filtered", "error", "normal"]
        priorities = args.priorities or [TestCase.EnumPriority.Normal, TestCase.EnumPriority.High]
        status = args.status or [TestCase.EnumStatus.Ready]

        test_conf = TestCaseSettings(names=args.tests,
                                    excluded_names=args.excluded_names,
                                    priorities=priorities,
                                    status=status,
                                    owners=args.owners,
                                    tags=args.tags,
                                    excluded_tags=args.excluded_tags)

        loader = TestLoader(test_conf.filter)
        tests = loader.load(test_conf.names)

        test_list_output = test_list_types[args.output_type](args.output_file)

        if "filtered" in shows:
            filtered_tests = loader.get_filtered_tests_with_reason().items()
            sorted_filtered_tests = sorted(filtered_tests, key=lambda x: x[0].test_name)
            test_list_output.output_filtered_tests(sorted_filtered_tests)

        if "normal" in shows:
            sorted_tests = sorted(tests, key=lambda x: x.test_name)
            test_list_output.output_normal_tests(sorted_tests)

        if "error" in shows:
            error_tests = loader.get_last_errors().items()
            sorted_error_tests = sorted(error_tests, key=lambda x: x[0])
            test_list_output.output_error_tests(sorted_error_tests)
        test_list_output.end_output()


class RunPlan(Command):
    """执行测试计划
    """
    name = "runplan"
    parser = argparse.ArgumentParser("Run QTA test plan")
    parser.add_argument("--report-type", help="report type", choices=report_types.keys(), default="stream")
    parser.add_argument("--report-args", help="additional arguments for specific report", default="")
    parser.add_argument("--report-args-help", help="show help information for specific report arguemnts", choices=report_types.keys())

    parser.add_argument("--runner-type", help="test runner type", choices=runner_types.keys(), default="basic")
    parser.add_argument("--runner-args", help="additional arguments for specific runner", default="")
    parser.add_argument("--runner-args-help", help="show help information for specific runner arguemnts", choices=runner_types.keys())

    parser.add_argument("--resmgr-backend-type", help="test resource manager backend type", choices=resmgr_backend_types.keys(), default="local")

    parser.add_argument("plan", help="designate a test plan to run")

    def execute(self, args):
        """执行过程
        """
        if args.report_args_help:
            report_types[args.report_args_help].get_parser().print_help()
            return
        if args.runner_args_help:
            runner_types[args.runner_args_help].get_parser().print_help()
            return

        report_type = report_types[args.report_type]
        report = report_type.parse_args(shlex.split(args.report_args))

        resmgr_backend = resmgr_backend_types[args.resmgr_backend_type]()

        runner_type = runner_types[args.runner_type]
        runner = runner_type.parse_args(shlex.split(args.runner_args), report, resmgr_backend)

        planname = args.plan
        planmodulename, planclsname = planname.rsplit(".", 1)
        __import__(planmodulename)
        planmod = sys.modules[planmodulename]
        plancls = getattr(planmod, planclsname)
        runner.run(plancls())
        if not report.is_passed():
            return 1


class InstallLib(Command):
    """安装扩展库
    """
    name = 'installlib'
    parser = argparse.ArgumentParser("Install extend library")
    parser.add_argument("egg_path", help="egg file path")

    def execute(self, args):
        """执行过程
        """
        ExLibManager(settings.PROJECT_ROOT).install(args.egg_path)


class CreateProject(Command):
    """创建测试项目
    """
    name = 'createproject'
    parser = argparse.ArgumentParser("Create new QTA project")
    parser.add_argument('name', help="project name")
    parser.add_argument('--dest', default=None, help="target path to create project")
    # parser.add_argument('--mode', default="standalone", help="project mode: standard or standalone")

    def execute(self, args):
        """执行过程
        """
        if args.dest is None:
            dest = os.getcwd()
        else:
            dest = args.dest
        if os.path.isfile(__file__):
            mode = project.EnumProjectMode.Standard
        else:
            mode = project.EnumProjectMode.Standalone
        proj_path = os.path.join(dest, args.name.lower() + 'testproj')
        logger.info('create %s mode test project: %s' % (mode, proj_path))
        project.create_project(proj_path, args.name, mode)


class UpgradeProject(Command):
    """升级测试项目
    """
    name = 'upgradeproject'
    parser = argparse.ArgumentParser("Upgrade QTA project")
    parser.add_argument('proj_path', help="target path to upgrade project")

    def execute(self, args):
        """执行过程
        """
        if os.path.isfile(__file__):
            logger.error("should run in egg mode")
            return 1

        qtaf_path = os.path.dirname(os.path.abspath(__file__))
        project.update_project_qtaf(args.proj_path, qtaf_path)


class Shell(Command):
    """Python Shell
    """
    name = 'shell'
    parser = argparse.ArgumentParser("open python shell")

    def execute(self, args):
        """执行过程
        """
        try:
            import readline  # optional, will allow Up/Down/History in the console
        except ImportError:
            pass
        import code
        varables = globals().copy()
        varables.update(locals())
        shell = code.InteractiveConsole(varables)
        shell.interact()


class Distribute(Command):
    """Generate distribution packages
    """
    name = 'dist'
    parser = argparse.ArgumentParser("Generate distribution packages")
    parser.add_argument("--version", help="assign distribution version", default="1.0.0")
    parser.add_argument("--exclude-resources", action="store_true", help="do not package resource data in package", default=False)

    def execute(self, args):
        DistGenerator(args.version).run(args.exclude_resources)


class RunTestDistPackage(Command):
    """Run tests in distribution package
    """
    name = "runtest"
    parser = argparse.ArgumentParser("Run tests in distribution package")
    parser.add_argument("--recreate-venv", action="store_true", help="force recreate virtualenv")
    parser.add_argument("--venv", help="designate virtualenv path manually")

    parser.add_argument("--priority", help="run test cases with specific priority, accept multiple options",
                        choices=[TestCasePriority.BVT, TestCasePriority.High, TestCasePriority.Normal, TestCasePriority.Low],
                        action="append", dest="priorities")
    parser.add_argument("--status", help="run test cases with specific status, accept multiple options",
                        choices=[TestCaseStatus.Design, TestCaseStatus.Implement, TestCaseStatus.Ready, TestCaseStatus.Review, TestCaseStatus.Suspend],
                        action="append", dest="status")
    parser.add_argument("--owner", help="run test cases with specific owner, accept multiple options", action="append", dest="owners")
    parser.add_argument("--excluded-name", help="exclude test cases with specific name prefix , accept multiple options", action="append", dest="excluded_names")
    parser.add_argument("--tag", help="run test cases with specific tag, accept multiple options", action="append", dest="tags")
    parser.add_argument("--excluded-tag", help="exclude test cases with specific name tag , accept multiple options", action="append", dest="excluded_tags")

    parser.add_argument("--report-type", help="report type", choices=report_types.keys(), default="stream")
    parser.add_argument("--report-args", help="additional arguments for specific report", default="")
    parser.add_argument("--report-args-help", help="show help information for specific report arguemnts", choices=report_types.keys())

    parser.add_argument("--runner-type", help="test runner type", choices=runner_types.keys(), default="basic")
    parser.add_argument("--runner-args", help="additional arguments for specific runner", default="")
    parser.add_argument("--runner-args-help", help="show help information for specific runner arguemnts", choices=runner_types.keys())

    parser.add_argument("--resmgr-backend-type", help="test resource manager backend type", choices=resmgr_backend_types.keys(), default="local")

    parser.add_argument("--disable-run-on-child", help="do not create child process to run", action="store_true")

    parser.add_argument("package", help="QTA sdist package")
    parser.add_argument("tests", nargs='*', help="designate test names to run")

    def execute(self, args):
        if args.report_args_help:
            report_types[args.report_args_help].get_parser().print_help()
            return
        if args.runner_args_help:
            runner_types[args.runner_args_help].get_parser().print_help()
            return

        venv = VirtuelEnv(
            args.package,
            args.venv,
            args.recreate_venv)
        venv.activate()

        report_type = report_types[args.report_type]
        report = report_type.parse_args(shlex.split(args.report_args))

        resmgr_backend = resmgr_backend_types[args.resmgr_backend_type]()

        runner_type = runner_types[args.runner_type]
        runner_args = args.runner_args + " "  # incase user input has no space
        runner = runner_type.parse_args(shlex.split(runner_args), report, resmgr_backend)

        test = TestCaseSettings(
                args.tests,
                args.excluded_names,
                args.priorities,
                args.status,
                args.owners,
                args.tags,
                args.excluded_tags)

        runner.run(test)
        if not report.is_passed():
            return 1


class RunPlanDistPackage(Command):
    """Run test plan in distribution package
    """
    name = "runplan"
    parser = argparse.ArgumentParser("Run test plan in distribution package")
    parser.add_argument("--recreate-venv", action="store_true", help="force recreate virtualenv")
    parser.add_argument("--venv", help="designate virtualenv path manually")

    parser.add_argument("--report-type", help="report type", choices=report_types.keys(), default="stream")
    parser.add_argument("--report-args", help="additional arguments for specific report", default="")
    parser.add_argument("--report-args-help", help="show help information for specific report arguemnts", choices=report_types.keys())

    parser.add_argument("--runner-type", help="test runner type", choices=runner_types.keys(), default="basic")
    parser.add_argument("--runner-args", help="additional arguments for specific runner", default="")
    parser.add_argument("--runner-args-help", help="show help information for specific runner arguemnts", choices=runner_types.keys())

    parser.add_argument("--resmgr-backend-type", help="test resource manager backend type", choices=resmgr_backend_types.keys(), default="local")

    parser.add_argument("--disable-run-on-child", help="do not create child process to run", action="store_true")

    parser.add_argument("package", help="QTA sdist package")
    parser.add_argument("plan", help="designate a test plan to run")

    def execute(self, args):
        if args.report_args_help:
            report_types[args.report_args_help].get_parser().print_help()
            return
        if args.runner_args_help:
            runner_types[args.runner_args_help].get_parser().print_help()
            return

        venv = VirtuelEnv(
            args.package,
            args.venv,
            args.recreate_venv)
        venv.activate()

        report_type = report_types[args.report_type]
        report = report_type.parse_args(shlex.split(args.report_args))

        resmgr_backend = resmgr_backend_types[args.resmgr_backend_type]()

        runner_type = runner_types[args.runner_type]
        runner = runner_type.parse_args(shlex.split(args.runner_args), report, resmgr_backend)

        planname = args.plan
        planmodulename, planclsname = planname.rsplit(".", 1)
        __import__(planmodulename)
        planmod = sys.modules[planmodulename]
        plancls = getattr(planmod, planclsname)
        runner.run(plancls())
        if not report.is_passed():
            return 1


class ManagementToolsConsole(object):
    """管理工具交互模式
    """
    prompt = "QTA> "

    def __init__(self, argparser):
        self._argparser = argparser

    def cmdloop(self):
        logger.info("""QTAF %(qtaf_version)s (test project: %(proj_root)s [%(proj_mode)s mode])\n""" % {
            'qtaf_version': version,
            'proj_root': settings.PROJECT_ROOT,
            'proj_mode': settings.PROJECT_MODE,
        })
        if six.PY3:
            raw_input_func = input
        else:
            raw_input_func = raw_input
        while 1:
            line = raw_input_func(self.prompt)
            args = shlex.split(line, posix="win" not in sys.platform)
            if not args:
                continue
            subcmd = args[0]
            if not self._argparser.get_subcommand(subcmd):
                sys.stderr.write("invalid command: \"%s\"\n" % subcmd)
                continue
            try:
                subcmd, ns = self._argparser.parse_args(args)
                subcmd.execute(ns)
            except SystemExit:
                logger.info("command exit")
            except:
                traceback.print_exc()


class ManagementTools(object):
    """管理工具类入口
    """
    excluded_command_types = [CreateProject, UpgradeProject, RunTestDistPackage, RunPlanDistPackage]

    def __init__(self):
        if settings.PROJECT_MODE == project.EnumProjectMode.Standard:
            self.excluded_command_types.append(InstallLib)

    def _load_cmds(self):
        """加载全部的命令
        """
        cmds = []
        cmds += self._load_cmd_from_module(sys.modules[__name__])
        cmds += self._load_app_cmds()
        return cmds

    def _load_cmd_from_module(self, mod):
        """加载一个模块里面的全部命令
        """
        cmds = []
        for objname in dir(mod):
            obj = getattr(mod, objname)
            if not inspect.isclass(obj):
                continue
            if obj == Command:
                continue
            if obj in self.excluded_command_types:
                continue
            if issubclass(obj, Command):
                cmds.append(obj)

        cmds.sort(key=lambda x: x.name)
        return cmds

    def _load_app_cmds(self):
        """从应用lib中加载命令
        """
        cmds = []
        for libname in settings.INSTALLED_APPS:
            if not libname:
                continue
            modname = '%s.cmds' % libname
            try:
                __import__(modname)
            except ImportError:
                continue
            else:
                for cmd in self._load_cmd_from_module(sys.modules[modname]):
                    cmd.name = libname + '.' + cmd.name
                    cmds.append(cmd)
        return cmds

    def run(self):
        """执行入口
        """
        cmds = self._load_cmds()
        argparser = ArgumentParser(cmds)
        if len(sys.argv) > 1:
            subcmd, args = argparser.parse_args(sys.argv[1:])
            return subcmd.execute(args) or 0
        else:
            ManagementToolsConsole(argparser).cmdloop()
            return 0


def qta_manage_main():
    """qta-manage工具入口
    """
    use_egg = not os.path.isfile(__file__)
    if use_egg:
        cmds = [CreateProject, UpgradeProject, RunTestDistPackage, RunPlanDistPackage, Help]
    else:
        cmds = [CreateProject, RunTestDistPackage, RunPlanDistPackage, Help]
    argparser = ArgumentParser(cmds)
    subcmd, args = argparser.parse_args(sys.argv[1:])
    exit_code = subcmd.execute(args) or 0
    os._exit(exit_code)
